@@ -18,12 +18,3 @@ module Bytes = struct
   let cat = (^)
 end
 let output_bytes = output_string
-
-module Buffer = struct
-  include Buffer
-
-  let to_bytes = contents
-end
-
-(* Introduced in 4.01.0 *)
-let ( |> ) x f = f x
