@@ -296,8 +296,7 @@ let rec dump ic pos len w =
     | '\032'..'\127' as c -> print_char c
     | _ -> print_char '.'
   done;
-  Printf.printf "\n";
-  flush stdout;
+  Printf.printf "\n%!";
   dump ic (pos + l) (len - l) w
 
 module Symbol = struct
@@ -344,7 +343,7 @@ module Symbol = struct
     }
 
   let is_extern = function
-    |  { storage = 2; section = `Num 0; value = 0l } -> true
+    | { storage = 2; section = `Num 0; value = 0l } -> true
     | _ -> false
 
   let is_export = function
@@ -481,7 +480,7 @@ end
 module Section = struct
   let create name flags = {
     sec_pos = (-1); sec_name = name; data = `String (Bytes.of_string ""); relocs = [];
-    vaddress = 0l; vsize = 0l;  sec_opts = flags;
+    vaddress = 0l; vsize = 0l; sec_opts = flags;
   }
 
   let nreloc_ovfl = 0x01000000l
@@ -736,7 +735,7 @@ module Coff = struct
                 s.auxs <- Bytes.make (Bytes.length auxs) '\000'
             | _ ->
                 Symbol.dump s;
-                Printf.printf "aux=%S\n" (Bytes.to_string s.auxs);
+                Printf.printf "aux=%S\n%!" (Bytes.to_string s.auxs);
                 assert false);
          (match s.section with
             | `Num i when i > 0 && i <= Array.length sections ->
@@ -773,7 +772,7 @@ module Coff = struct
     Printf.printf "opts:    0x%x\n" x.opts;
     List.iter Symbol.dump x.symbols;
     List.iter Section.dump x.sections;
-    ()
+    flush stdout
 
   let put oc x =
     emit_int16 oc x.machine;
@@ -869,7 +868,7 @@ module Lib = struct
       let buf = read ic (pos_in ic) 60 in
       let base = pos_in ic in
       let size = int_of_string (strz (Bytes.sub buf 48 10) 0 ' ') in
-      let name = strz (Bytes.sub buf 0 16) 0 ' '  in
+      let name = strz (Bytes.sub buf 0 16) 0 ' ' in
       begin match name with
         | "/" | "" -> ()
         | "//" -> strtbl := read ic (pos_in ic) size
