@@ -13,6 +13,13 @@
 open Coff
 open Cmdline
 
+let debug ?(dry_mode = {contents=false}) min_level fmt =
+  let print msg =
+    if !dry_mode || !verbose >= min_level then
+      Printf.printf "%s\n%!" msg
+  in
+    Printf.ksprintf print fmt
+
 let search_path = ref []
 let default_libs = ref []
 
@@ -64,11 +71,16 @@ let read_file fn =
   close_in ic;
   List.rev !r
 
+(* This is the longest command which can be passed to [Sys.command] *)
+let max_command_length =
+  let processor = try Sys.getenv "COMSPEC" with Not_found -> "cmd.exe" in
+  (* The 4 is from the " /c " *)
+  8191 - String.length processor - 4
 
 let get_output ?(use_bash = false) ?(accept_error=false) cmd =
   let fn = Filename.temp_file "flexdll" "" in
   let cmd' = cmd ^ " > " ^ (Filename.quote fn) in
-    if String.length cmd' < 8182 && not use_bash then
+    if String.length cmd' < max_command_length && not use_bash then
       begin
         if (Sys.command cmd' <> 0) && not accept_error
         then failwith ("Cannot run " ^ cmd);
@@ -168,11 +180,8 @@ let build_diversion lst =
       ) (List.filter (fun f -> f <> "") lst)
   in
   let utf16, lst =
-    match List.map toutf16 lst with
-    | lst ->
-        true, lst
-    | exception Not_utf8 ->
-        false, lst
+    try true, List.map toutf16 lst
+    with Not_utf8 -> false, lst
   in
   if utf16 then output_string oc "\xFF\xFE"; (* LE BOM *)
   List.iter (fun s -> output_string oc s) lst;
@@ -186,7 +195,7 @@ type cmdline = {
 let new_cmdline () =
   let rf = match !toolchain with
   | `MSVC | `MSVC64 | `LIGHTLD -> true
-  | `MINGW | `MINGW64 | `GNAT | `CYGWIN | `CYGWIN64 -> false
+  | `MINGW | `MINGW64 | `GNAT | `GNAT64 | `CYGWIN | `CYGWIN64 -> false
   in
   {
    may_use_response_file = rf;
@@ -201,14 +210,14 @@ let run_command cmdline cmd =
   (* note: for Cygwin, using bash allow to follow symlinks to find
      gcc... *)
   if !toolchain = `CYGWIN || !toolchain = `CYGWIN64 ||
-     String.length cmd_quiet >= 8192
+     String.length cmd_quiet >= max_command_length
   then begin
     (* Dump the command in a text file and apply bash to it. *)
     let (fn, oc) = open_temp_file "longcmd" "" in
     output_string oc cmd;
     close_out oc;
 
-    if !verbose >= 1 then Printf.printf "(call with bash: %s)\n%!" fn;
+    debug 1 "(call with bash: %s)" fn;
     if Sys.command (Printf.sprintf "bash %s" fn) <> 0 then
       failwith "Error during linking\n"
   end else
@@ -312,7 +321,7 @@ let int_to_buf b i =
 let exportable s =
   match !machine with
   | `x86 ->
-      s <> "" && (s.[0] = '_'  || s.[0] = '?')
+      s <> "" && (s.[0] = '_' || s.[0] = '?')
   | `x64 ->
       if String.length s > 2 && s.[0] = '?' && s.[1] = '?' then false
       else true
@@ -389,17 +398,17 @@ let add_reloc_table obj obj_name p =
         | `x64, 0x08 when not !no_rel_relocs-> 0x0003 (* rel32_4 *)
         | `x64, 0x06 when not !no_rel_relocs-> 0x0005 (* rel32_2 *)
 
-        | `x86, (0x0a (* IMAGE_REL_I386_SECTION *) |
-                 0x0b (* IMAGE_REL_I386_SECREL*) ) ->
+        | (`x86 | `x64), (0x0a (* IMAGE_REL_{I386|AMD64}_SECTION *) |
+                          0x0b (* IMAGE_REL_{I386|AMD64}_SECREL*) ) ->
             0x0100 (* debug relocs: ignore *)
 
-        | _, k  ->
+        | _, k ->
             let msg =
               Printf.sprintf "Unsupported relocation kind %04x for %s in %s"
                 k rel.symbol.sym_name obj_name
             in
             failwith msg
-(*            Printf.eprintf "%s\n" msg;
+(*            Printf.eprintf "%s\n%!" msg;
             0x0001 *)
       in
       int_to_buf data kind;
@@ -456,7 +465,7 @@ let add_reloc_table obj obj_name p =
       Reloc.abs !machine sect (Int32.of_int (Buffer.length data)) secsym;
       int_to_buf data (Int32.to_int min);
       Reloc.abs !machine sect (Int32.of_int (Buffer.length data)) secsym;
-      int_to_buf data (Int32.to_int  max);
+      int_to_buf data (Int32.to_int max);
       int_to_buf data 0;
     )
     !nonwr;
@@ -498,11 +507,11 @@ let add_export_table obj exports symname =
   let strings = Buffer.create 1024 in
   let strsym = Symbol.intern sect 0l in
   obj.symbols <- strsym :: (Symbol.export symname sect 0l) :: obj.symbols;
-  let exports = List.sort Pervasives.compare exports in
+  let exports = List.sort Stdlib.compare exports in
   (* The runtime library assumes that names are sorted! *)
   int_to_buf data (List.length exports);
   List.iter
-    (fun s  ->
+    (fun s ->
        let sym = Symbol.extern s in
        obj.symbols <- sym :: obj.symbols;
        Reloc.abs !machine sect (Int32.of_int (Buffer.length data)) sym;
@@ -527,7 +536,7 @@ let add_master_reloc_table obj names symname =
   let data = Buffer.create 1024 in
   obj.symbols <- (Symbol.export symname sect 0l) :: obj.symbols;
   List.iter
-    (fun s  ->
+    (fun s ->
        let sym = Symbol.extern s in
        obj.symbols <- sym :: obj.symbols;
        Reloc.abs !machine sect (Int32.of_int (Buffer.length data)) sym;
@@ -561,7 +570,7 @@ let collect f l =
     l
 
 let cmd_verbose cmd =
-  if !verbose >= 1 then Printf.printf "+ %s\n" cmd;
+  debug 1 "+ %s" cmd;
   Sys.command cmd
 
 
@@ -585,7 +594,7 @@ let parse_dll_exports fn =
 let dll_exports fn = match !toolchain with
   | `MSVC | `MSVC64 | `LIGHTLD ->
       failwith "Creation of import library not supported for this toolchain"
-  | `GNAT | `CYGWIN | `CYGWIN64 | `MINGW | `MINGW64 ->
+  | `GNAT | `GNAT64 | `CYGWIN | `CYGWIN64 | `MINGW | `MINGW64 ->
       let dmp = temp_file "dyndll" ".dmp" in
       if cmd_verbose (Printf.sprintf "%s -p %s > %s" !objdump fn dmp) <> 0
       then failwith "Error while extracting exports from a DLL";
@@ -605,7 +614,7 @@ let patch_output filename =
       in
       begin try Stacksize.set_stack_reserve filename x
       with exn ->
-        Printf.eprintf "Cannot set stack reserve: %s"
+        Printf.eprintf "Cannot set stack reserve: %s\n%!"
           (Printexc.to_string exn)
       end
   | None -> ()
@@ -629,6 +638,9 @@ let needed imported defined unalias obj =
     | Some s when not (StrSet.mem name defined) ->
         imported := StrSet.add s !imported;
         if StrSet.mem s defined then s else name
+    | None when not (StrSet.mem name defined) && StrSet.mem ("__imp_" ^ name) defined ->
+        imported := StrSet.add ("__imp_" ^ name) !imported;
+        "__imp_" ^ name
     | _ -> name
   in
   List.fold_left
@@ -657,7 +669,10 @@ let build_dll link_exe output_file files exts extra_args =
   let loaded_filenames : (string,unit) Hashtbl.t = Hashtbl.create 16 in
   let read_file fn =
     if Lib.is_dll fn then `Lib ([], dll_exports fn)
-    else Lib.read fn
+    else begin
+      if !verbose >= 2 then Printf.printf "** open: %s\n" fn;
+      Lib.read fn
+    end
   in
   let files = List.map (fun fn -> fn, read_file fn) files in
 
@@ -666,6 +681,16 @@ let build_dll link_exe output_file files exts extra_args =
   let objs = collect (function (f, `Obj x) -> Some (f,x) | _ -> None) files in
   let libs = collect (function (f, `Lib (x,_)) -> Some (f,x) | _ -> None) files in
 
+  let with_data_symbol symbols sym_name f =
+    if !toolchain <> `MSVC && !toolchain <> `MSVC64 then
+      match check_prefix "__nm_" sym_name with
+      | None -> ()
+      | Some s ->
+        let imp_name = "__imp_" ^ s  in
+        if List.exists ( fun p -> Symbol.is_defin p &&
+          p.sym_name = imp_name) symbols then
+          f s;
+  in
   (* Collect all the available symbols, including those defined
      in default libraries *)
   let defined, from_imports, unalias =
@@ -690,7 +715,7 @@ let build_dll link_exe output_file files exts extra_args =
       (* Collect aliases *)
       List.iter
         (fun (x, y) ->
-           if !verbose >= 2 then Printf.printf "alias %s -> %s\n" x y;
+           debug 2 "alias %s -> %s" x y;
            Hashtbl.add aliases x y
         )
         (Coff.aliases obj);
@@ -698,12 +723,12 @@ let build_dll link_exe output_file files exts extra_args =
 
       (* Iterates through DEFAULTLIB directives *)
       let register_deflib fn =
-        if not (is_crt_lib fn) then
+        if not !custom_crt || not (is_crt_lib fn) then
           let fn = find_file fn in
           if not (Hashtbl.mem loaded_filenames fn)
           then (Hashtbl.add loaded_filenames fn (); collect_file fn)
       in
-      if not !builtin_linker && !use_default_libs && not !custom_crt then
+      if not !builtin_linker && !use_default_libs then
         List.iter
           (fun (cmd, args) ->
              if String.uppercase_ascii cmd = "DEFAULTLIB" then List.iter register_deflib args
@@ -712,13 +737,17 @@ let build_dll link_exe output_file files exts extra_args =
 
       (* Collect defined symbols *)
       List.iter
-        (fun sym -> if Symbol.is_defin sym then add_def sym.sym_name)
+        (fun sym ->
+          if Symbol.is_defin sym then (
+            add_def sym.sym_name;
+            with_data_symbol obj.symbols sym.sym_name add_def)
+        )
         obj.symbols
 
     and collect_file fn =
       if not (Hashtbl.mem collected (String.lowercase_ascii fn)) then begin
         Hashtbl.replace collected (String.lowercase_ascii fn) ();
-        if !verbose >= 2 then Printf.printf "** open: %s\n" fn;
+        debug 2 "** open: %s" fn;
         collect_defined fn (Lib.read fn)
       end
 
@@ -728,8 +757,7 @@ let build_dll link_exe output_file files exts extra_args =
           List.iter (fun (_, obj) -> collect_defined_obj obj) objs;
           List.iter
             (fun (s,_) ->
-               if !verbose >= 2 then
-                 Printf.printf "lib %s import symbol %s\n%!" fn s;
+               debug 2 "lib %s import symbol %s" fn s;
                from_imports := StrSet.add s !from_imports;
                add_def s;
                add_def ("__imp_" ^ s)
@@ -764,10 +792,13 @@ let build_dll link_exe output_file files exts extra_args =
         (fun sym ->
            if Symbol.is_defin sym
            then begin
-             if !explain then
-               Printf.printf "Symbol %s found in %s(%s)\n%!" sym.sym_name fn
-                 objname;
-             Hashtbl.replace defined_in sym.sym_name (fn, objname, obj);
+             let f s =
+               if !explain then
+                 Printf.printf "Symbol %s found in %s(%s)\n%!" s fn objname;
+               Hashtbl.replace defined_in s (fn,objname,obj);
+             in
+             f sym.sym_name;
+             with_data_symbol obj.symbols sym.sym_name f
            end
         )
         obj.symbols
@@ -799,7 +830,7 @@ let build_dll link_exe output_file files exts extra_args =
 
   (* Second step: transitive closure, starting from given objects *)
 
-  let libobjects  = Hashtbl.create 16 in
+  let libobjects = Hashtbl.create 16 in
   let reloctbls = ref [] in
   let exported = ref StrSet.empty in
 
@@ -818,7 +849,7 @@ let build_dll link_exe output_file files exts extra_args =
 
   let add_reloc name obj imps =
     if !show_imports && not (StrSet.is_empty imps) then (
-      Printf.printf "** Imported symbols for %s:\n" name;
+      Printf.printf "** Imported symbols for %s:\n%!" name;
       StrSet.iter print_endline imps
     );
     let sym = add_reloc_table obj name (fun s -> StrSet.mem s.sym_name imps) in
@@ -890,7 +921,7 @@ let build_dll link_exe output_file files exts extra_args =
           (* the linker will find this object in this library *)
       else begin
         if !explain then
-          Printf.printf "Library object %s(%s) needs to be rewritten\n"
+          Printf.printf "Library object %s(%s) needs to be rewritten\n%!"
             libname objname;
         Hashtbl.add redirect libname
           (close_obj (Printf.sprintf "%s(%s)" libname objname) imps obj)
@@ -903,6 +934,7 @@ let build_dll link_exe output_file files exts extra_args =
     StrSet.iter print_endline !exported;
     Printf.printf "** Symbols from import libs:\n";
     StrSet.iter print_endline !imported_from_implib;
+    flush stdout
   );
 
   if !reexport_from_implibs then
@@ -913,7 +945,7 @@ let build_dll link_exe output_file files exts extra_args =
 
   if not (StrSet.is_empty !imported) then begin
 (*
-    Printf.printf "** __imp symbols:\n";
+    Printf.printf "** __imp symbols:\n%!";
     StrSet.iter print_endline !imported;
 *)
     add_import_table obj (StrSet.elements !imported);
@@ -976,7 +1008,7 @@ let build_dll link_exe output_file files exts extra_args =
           if x = "" then Filename.chop_extension output_file ^ ".def"
           else x
         in
-        if !verbose >= 1 then Printf.printf "Generate %s\n%!" fn;
+        debug 1 "Generate %s" fn;
         let oc = open_out fn in
         Printf.fprintf oc "LIBRARY %s\n" output_file;
         Printf.fprintf oc "EXPORTS\n";
@@ -1043,9 +1075,6 @@ let build_dll link_exe output_file files exts extra_args =
             close_out oc;
             Filename.quote def_file
         in
-        let extra_args =
-          if !machine = `x64 then "-Xlinker --image-base -Xlinker 0x10000 " ^ extra_args else extra_args
-        in
         Printf.sprintf
           "%s %s%s -L. %s %s -o %s %s %s %s %s"
           !gcc
@@ -1058,7 +1087,7 @@ let build_dll link_exe output_file files exts extra_args =
           files
           def_file
           extra_args
-    | `MINGW | `MINGW64 | `GNAT ->
+    | `MINGW | `MINGW64 | `GNAT | `GNAT64 ->
         let def_file =
           if main_pgm then ""
           else
@@ -1093,7 +1122,7 @@ let build_dll link_exe output_file files exts extra_args =
           (if !implib then "--out-implib " ^ Filename.quote (Filename.chop_extension output_file ^ ".a") else "")
           extra_args
   in
-  if !verbose >= 1 || !dry_mode then Printf.printf "+ %s\n%!" cmd;
+  debug ~dry_mode 1 "+ %s" cmd;
   if not !dry_mode then begin
     let manifest_file = output_file ^ ".manifest" in
     safe_remove manifest_file;
@@ -1117,7 +1146,7 @@ let build_dll link_exe output_file files exts extra_args =
                            else output_file ^ ";#2"))
           (Filename.quote fn)
       in
-      if !verbose >= 1 then Printf.printf "+ %s\n%!" mcmd;
+      debug 1 "+ %s" mcmd;
       if Sys.command mcmd <> 0 then
         failwith "Error while merging the manifest";
       safe_remove manifest_file;
@@ -1217,12 +1246,13 @@ let setup_toolchain () =
       |> remove_duplicate_paths
     in
     search_path := !dirs @ lib_search_dirs;
-    if !verbose >= 1  then begin
+    if !verbose >= 1 then begin
       print_endline "lib search dirs:";
-      List.iter (Printf.printf "  %s\n%!") lib_search_dirs;
+      List.iter (Printf.printf "  %s\n") lib_search_dirs;
+      flush stdout
     end;
     default_libs :=
-      ["-lmingw32"; "-lgcc"; "-lmoldname"; "-lmingwex"; "-lmsvcrt";
+      ["-lmingw32"; "-lgcc"; "-lgcc_eh"; "-lmoldname"; "-lmingwex"; "-lmsvcrt";
        "-luser32"; "-lkernel32"; "-ladvapi32"; "-lshell32" ];
     if !exe_mode = `EXE then default_libs := "crt2.o" :: !default_libs
     else default_libs := "dllcrt2.o" :: !default_libs
@@ -1253,7 +1283,7 @@ let setup_toolchain () =
       mingw_libs Version.mingw_prefix
   | `MINGW64 ->
       mingw_libs Version.mingw64_prefix
-  | `GNAT ->
+  | `GNAT | `GNAT64 ->
    (* This is a plain copy of the mingw version, but we do not change the
       prefix and use "gnatls" to compute the include dir. *)
     search_path :=
@@ -1309,7 +1339,7 @@ let compile_if_needed file =
             (Filename.quote tmp_obj)
             (mk_dirs_opt "-I")
             file
-      | `MINGW | `MINGW64 | `GNAT ->
+      | `MINGW | `MINGW64 | `GNAT | `GNAT64 ->
           Printf.sprintf
             "%s -c -o %s %s %s"
             !gcc
@@ -1319,7 +1349,7 @@ let compile_if_needed file =
       | `LIGHTLD ->
           failwith "Compilation of C code is not supported for this toolchain"
     in
-    if !verbose >= 1 || !dry_mode then Printf.printf "+ %s\n%!" cmd;
+    debug ~dry_mode 1 "+ %s" cmd;
     let exit = if !dry_mode then 0 else Sys.command cmd in
     if pipe <> "" then display_msvc_output stdout file;
     if exit <> 0 then failwith "Error while compiling";
@@ -1340,7 +1370,8 @@ let dump fn =
         objs;
       List.iter
         (fun (s,i) -> Printf.printf "** import: %s (%i)\n" s i)
-        imports
+        imports;
+      flush stdout
   | `Obj o ->
       Coff.dump o
 
@@ -1359,7 +1390,8 @@ let all_files () =
   | `CYGWIN -> "cygwin.o"
   | `CYGWIN64 -> "cygwin64.o"
   | `MINGW64 -> "mingw64.o"
-  | `GNAT    -> "gnat.o"
+  | `GNAT -> "gnat.o"
+  | `GNAT64 -> "gnat64.o"
   | `MINGW | `LIGHTLD -> "mingw.o" in
   if !exe_mode <> `DLL then
     if !add_flexdll_obj then f ("flexdll_" ^ tc) :: files
@@ -1377,7 +1409,7 @@ let main () =
       match !toolchain, !cygpath_arg with
       | _, `Yes -> true
       | _, `No -> false
-      | (`GNAT|`MINGW|`MINGW64|`CYGWIN|`CYGWIN64), `None ->
+      | (`GNAT|`GNAT64|`MINGW|`MINGW64|`CYGWIN|`CYGWIN64), `None ->
           begin match Sys.os_type with
           | "Unix" | "Cygwin" ->
               Sys.command "cygpath -S 2>/dev/null >/dev/null" = 0
@@ -1396,7 +1428,8 @@ let main () =
     if !use_default_libs then begin
       Printf.printf "** Default libraries:\n";
       List.iter print_endline !default_libs;
-    end
+    end;
+    flush stdout
    );
   let files = all_files () in
   match !mode with
